from __future__ import annotations

import eznet
from eznet.data import Data


from .system import System
from .chassis import Chassis
from .interfaces import Interface


class Device:
    def __init__(self, device: eznet.device.Device):
        self.system = System(device)
        self.chassis = Chassis(device)
        self.interfaces = Data(Interface.fetch, device)



import logging
from pathlib import Path
from typing import Union

import asyncssh

SFTP_HOST = 'sftp.juniper.net'
SFTP_USER = "jtac"
SFTP_PASS = "anonymous"


logger = logging.getLogger(__name__)


async def upload(path: Union[str, Path], case: str) -> None:
    if isinstance(path, str):
        path = Path(path)
    async with asyncssh.connect(SFTP_HOST, username=SFTP_USER, password=SFTP_PASS) as conn:
        logger.info("sftp: connected")
        async with conn.start_sftp_client() as sftp:
            if not await sftp.exists(f"/pub/incoming/{case}"):
                await sftp.mkdir(f"/pub/incoming/{case}")
            if not path.is_dir():
                logger.info(f"sftp: {path} --> {path.name}")
                await sftp.put(path, f"/pub/incoming/{case}")
                logger.info("sftp: done")
            else:
                for file in path.glob("**/*"):
                    if not file.is_dir():
                        remote_name = str(file.relative_to(path)).replace('/', '.')
                        logger.info(f"sftp: {file} --> {remote_name}")
                        await sftp.put(file, f"/pub/incoming/{case}/{remote_name}")
                        logger.info("sftp: done")



from __future__ import annotations

from datetime import datetime
from typing import Dict, Any, Optional, List

from attrs import define
from lxml.etree import _Element  # noqa

from eznet.parser.xml import text, number, timestamp
import eznet
from eznet.data import Data


@define
class Alarm:
    ts: Optional[datetime]
    cls: Optional[str]
    description: Optional[str]
    type: Optional[str]

    @staticmethod
    def from_xml(alarm: _Element) -> Alarm:
        return Alarm(
            ts=timestamp(alarm, "alarm-time"),
            cls=text(alarm, "alarm-class"),
            description=text(alarm, "alarm-description"),
            type=text(alarm, "alarm-type"),
        )

    @staticmethod
    async def fetch(device: eznet.device.Device) -> Optional[List[Alarm]]:
        xml = await device.junos.run_xml_cmd(
            "show chassis alarms",
        )
        if xml is not None:
            alarm_info = xml.find("alarm-information")
            if alarm_info is not None:
                return [
                    Alarm.from_xml(alarm)
                    for alarm in alarm_info.findall("alarm-detail")
                ]


@define
class Port:
    cable_type: Optional[str]
    fiber_mode: Optional[str]
    wavelength: Optional[str]

    @staticmethod
    def from_xml(port: _Element) -> Port:
        return Port(
            cable_type=text(port, "cable-type"),
            fiber_mode=text(port, "fiber-mode"),
            wavelength=text(port, "wavelength"),
        )


@define
class PIC:
    state: Optional[str]
    type: Optional[str]
    ports: Optional[Dict[int, Port]] = None

    @staticmethod
    def from_xml(pic: _Element) -> PIC:
        return PIC(
            state=text(pic, "pic-state"),
            type=text(pic, "pic-type"),
        )


@define
class FPC:
    state: Optional[str]
    comment: Optional[str]
    cpu_utilization_total: Optional[int]
    cpu_utilization_interrupt: Optional[int]
    memory_dram: Optional[int]
    memory_heap_utilization: Optional[int]
    memory_buffer_utilization: Optional[int]
    description: Optional[str]
    pics: Dict[int, PIC]

    @staticmethod
    def from_xml(fpc: _Element) -> FPC:
        return FPC(
            state=text(fpc, "state"),
            comment=text(fpc, "comment"),
            cpu_utilization_total=number(fpc, "cpu-total"),
            cpu_utilization_interrupt=number(fpc, "cpu-interrupt"),
            memory_dram=number(fpc, "memory-dram-size"),
            memory_heap_utilization=number(fpc, "memory-heap-utilization"),
            memory_buffer_utilization=number(fpc, "memory-buffer-utilization"),
            description=text(fpc, "description"),
            pics={
                number(pic, "pic-slot"): PIC.from_xml(pic)
                for pic in fpc.findall("pic")
            },
        )

    @staticmethod
    async def fetch(device: eznet.device.Device, get_ports: bool = False) -> Optional[Dict[int, FPC]]:
        show_chassis_fpc = await device.junos.run_xml_cmd("show chassis fpc pic-status")
        if show_chassis_fpc is not None:
            fpc_info = show_chassis_fpc.find("fpc-information")
            if fpc_info is not None:
                fpc_dict = {
                    number(fpc, "slot"): FPC.from_xml(fpc)
                    for fpc in fpc_info.findall("fpc")
                }

                if not get_ports:
                    return fpc_dict
                else:
                    for fpc_number, fpc in fpc_dict.items():
                        for pic_number, pic in fpc.pics.items():
                            xml = await device.junos.run_xml_cmd(
                                f"show chassis pic fpc-slot {fpc_number} pic-slot {pic_number}",
                            )
                            if xml is not None:
                                pic.ports = {
                                    number(port, "port-number"): Port.from_xml(port)
                                    for port in xml.findall("fpc-information/fpc/pic-detail/port-information/port")
                                }
                    return fpc_dict


@define
class FW:
    fw: Dict[str, str]

    @staticmethod
    def from_xml(xml: _Element) -> FW:
        fw = {
            text(e, "type"): text(e, "firmware-version")
            for e in xml.findall("firmware")
        }
        if "ONIE/DIAG" in fw:
            try:
                fw["ONIE"], fw["DIAG"] = fw.pop("ONIE/DIAG").split("/")
            except ValueError:
                pass
        fw = {key: value.strip() for key, value in fw.items() if value is not None}

        return FW(
            fw=fw,
        )

    @staticmethod
    async def fetch(device: eznet.device.Device) -> Optional[Dict[int, FW]]:
        show_chassis_fw = await device.junos.run_xml_cmd("show chassis firmware")
        if show_chassis_fw is not None:
            fw_info = show_chassis_fw.find("firmware-information/chassis")
            if fw_info is not None:
                return {
                    int(text(e, "name")[4:]):
                    FW.from_xml(e)
                    for e in fw_info.findall("chassis-module")
                    if "FPC " in text(e, "name")
                }


@define
class RE:
    model: Optional[str]
    status: Optional[str]
    mastership: Optional[str]
    start_time: Optional[datetime]
    reboot_reason: Optional[str]

    @staticmethod
    def from_xml(xml: _Element) -> RE:
        return RE(
            model=text(xml, "model"),
            status=text(xml, "status"),
            mastership=text(xml, "mastership-state"),
            start_time=timestamp(xml, "start-time"),
            reboot_reason=text(xml, "last-reboot-reason"),
        )

    @staticmethod
    async def fetch(device: eznet.device.Device) -> Optional[Dict[int, RE]]:
        show_chassis_re = await device.junos.run_xml_cmd("show chassis routing-engine")
        if show_chassis_re is not None:
            re_information = show_chassis_re.find("route-engine-information")
            if re_information is not None:
                return {
                    number(e, "slot"):
                    RE.from_xml(e) for e in re_information.findall("route-engine")
                    if number(e, "slot") is not None
                }


class Chassis:
    def __init__(self, device: eznet.device.Device):
        self.fpc = Data(FPC.fetch, device)
        self.re = Data(RE.fetch, device)
        self.fw = Data(FW.fetch, device)



from __future__ import annotations

from typing import List, Dict

from attrs import define
import pytest

import eznet
from eznet.data import Data


@define
class Info:
    value: str

    @classmethod
    async def fetch(cls, _: eznet.device.Device) -> Info:
        return Info(value="single")

    @classmethod
    async def fetch_list(cls, _: eznet.device.Device) -> List[Info]:
        return [
            Info(value="first"),
            Info(value="second"),
        ]

    @classmethod
    async def fetch_dict(cls, _: eznet.device.Device) -> Dict[str, Info]:
        return {
            "first": Info(value="first"),
            "second": Info(value="second"),
        }


class Model:
    def __init__(self) -> None:
        self.info = Data(Info.fetch, eznet.device.Device("test"))
        self.info_list = Data(Info.fetch_list, eznet.device.Device("test"))
        self.info_dict = Data(Info.fetch_dict, eznet.device.Device("test"))


@pytest.mark.asyncio
async def test_fetch():
    model = Model()
    await model.info.fetch("last")
    assert model.info["last"].value == "single"


@pytest.mark.asyncio
async def test_fetch_list():
    model = Model()
    await model.info_list.fetch("last")
    assert model.info_list["last"][0].value == "first"
    assert model.info_list["last"][1].value == "second"


@pytest.mark.asyncio
async def test_fetch_dict():
    model = Model()
    await model.info_dict.fetch("last")
    assert model.info_dict["last"]["first"].value == "first"
    assert model.info_dict["last"]["second"].value == "second"


def test_import():
    model = Model()
    model.info.imp0rt({"value": "single"}, "last")
    assert model.info["last"].value == "single"


def test_import_list():
    model = Model()
    model.info_list.imp0rt([{"value": "first"}, {"value": "second"}], "last")
    assert model.info_list["last"][0].value == "first"
    assert model.info_list["last"][1].value == "second"


def test_import_dict():
    model = Model()
    model.info_dict.imp0rt({"first": {"value": "first"}, "second": {"value": "second"}}, "last")
    assert model.info_dict["last"]["first"].value == "first"
    assert model.info_dict["last"]["second"].value == "second"


def test_export():
    model = Model()
    model.info.data["last"] = Info(value="single")
    assert model.info.exp0rt("last")["value"] == "single"


def test_export_list():
    model = Model()
    model.info_list.data["last"] = [Info(value="first"), Info(value="second")]
    assert model.info_list.exp0rt("last") == [{"value": "first"}, {"value": "second"}]


def test_export_dict():
    model = Model()
    model.info_dict.data["last"] = {"first": Info(value="first"), "second": Info(value="second")}
    assert model.info_dict.exp0rt("last") == {"first": {"value": "first"}, "second": {"value": "second"}}





from __future__ import annotations

import logging
from typing import List, Dict, Any, Optional, Union
from pathlib import Path
import json

import yaml
import _jsonnet

from eznet.device import Device

logger = logging.getLogger(__name__)


class Inventory:
    def __init__(self) -> None:
        self.devices: List[Device] = []

    def device(self, name: str) -> Optional[Device]:
        for device in self.devices:
            if device.name == name:
                return device
        else:
            return None

    def load(self, path: Union[str, Path]) -> Inventory:
        if isinstance(path, str):
            path = Path(path)
        path = path.expanduser()
        if not path.exists():
            logger.error(f"{path} not found")
        elif path.is_dir():
            logger.info(f"inventory: load from {path}/")
            for child in path.glob("*"):
                if child.is_dir() or child.suffix in [".yaml"]:
                    self.load(child)
        elif path.suffix == ".yaml":
            logger.info(f"inventory: load from {path}")
            try:
                with open(path) as io:
                    self.imp0rt(
                        yaml.safe_load(io.read()) or {},
                        site=path.with_suffix("").name,
                    )
            except Exception as exc:
                logger.error(f"inventory: load from {path}: {exc.__class__.__name__}: {exc}")
        elif path.suffix == ".json":
            logger.info(f"inventory: load from {path}")
            try:
                with open(path) as io:
                    self.imp0rt(
                        json.loads(io.read()),
                        site=path.with_suffix("").name,
                    )
            except Exception as exc:
                logger.error(f"inventory: load from {path}: {exc.__class__.__name__}: {exc}")
        elif path.suffix == ".jsonnet":
            logger.info(f"inventory: load from {path}")
            try:
                self.imp0rt(
                    json.loads(_jsonnet.evaluate_file(f"{path}")),
                    site=path.with_suffix("").name,
                )
            except Exception as exc:
                logger.error(f"inventory: load from {path}: {exc.__class__.__name__}: {exc}")
        else:
            logger.error(f"unknown inventory file format {path.suffix[1:]}")

        return self

    def imp0rt(self, data: Dict[str, Any], site: Optional[str] = None) -> None:
        devices: List[Dict[Any, Any]] = data.get("devices", [])
        if not isinstance(devices, list):
            return
        for device_data in devices:
            device_data.setdefault("site", site)
            device = Device(**device_data)
            if device in self.devices:
                logger.error(f"Load error: Duplicate device with {device.id}")
            else:
                self.devices.append(device)

    @property
    def sites(self) -> Dict[Union[str, None], List[Device]]:
        sites: Dict[Union[str, None], List[Device]] = {}
        for device in self.devices:
            if device.site not in sites:
                sites[device.site] = []
            sites[device.site].append(device)
        return sites



from __future__ import annotations

from typing import Optional, Dict, List
from datetime import datetime

from attrs import define
from lxml.etree import _Element  # noqa

import eznet
from eznet.parser.xml import text, timestamp, number
from eznet.data import Data


@define
class Info:
    hostname: Optional[str]
    sw_family: Optional[str]
    sw_version: Optional[str]
    hw_model: Optional[str]
    hw_sn: Optional[str]

    @staticmethod
    def from_xml(system_info: _Element) -> Info:
        return Info(
            hostname=text(system_info, "host-name"),
            sw_family=text(system_info, "os-name"),
            sw_version=text(system_info, "os-version"),
            hw_model=text(system_info, "hardware-model"),
            hw_sn=text(system_info, "serial-number"),
        )

    @staticmethod
    async def fetch(device: eznet.device.Device) -> Optional[Info]:
        xml = await device.junos.run_xml_cmd(
            "show system information",
        )
        if xml is not None:
            system_info = xml.find("system-information")
            if system_info is not None:
                return Info.from_xml(system_info)
        return None


@define
class SW:
    hostname: Optional[str]
    product_model: Optional[str]
    product_name: Optional[str]
    junos: Optional[str]

    @staticmethod
    def from_xml(soft_info: _Element) -> SW:
        return SW(
            hostname=text(soft_info, "host-name"),
            product_model=text(soft_info, "product-model"),
            product_name=text(soft_info, "product-name"),
            junos=text(soft_info, "junos-version"),
        )

    @staticmethod
    async def fetch(device: eznet.device.Device, both_re: bool = False) -> Optional[Dict[str, SW]]:
        if not both_re:
            xml = await device.junos.run_xml_cmd(
                "show version",
            )
        else:
            xml = await device.junos.run_xml_cmd(
                "show version invoke-on all-routing-engines",
            )

        if xml is not None:
            soft_info = xml.find("software-information")
            if soft_info is not None:
                return {
                    "localre": SW.from_xml(soft_info)
                }
            mre_results = xml.find("multi-routing-engine-results")
            if mre_results is not None:
                return {
                    text(mre_item, "re-name"):
                        SW.from_xml(mre_item.find("software-information"))
                    for mre_item in mre_results.findall("multi-routing-engine-item")
                    if text(mre_item, "re-name") is not None
                    and mre_item.find("software-information") is not None
                }
        return None


@define
class Uptime:
    current_time: Optional[datetime]
    system_boot_time: Optional[datetime]
    protocol_start_time: Optional[datetime]
    last_config_time: Optional[datetime]
    time_source: Optional[str]

    @staticmethod
    def from_xml(uptime_info: _Element) -> Uptime:
        return Uptime(
            current_time=timestamp(uptime_info, "current-time/date-time"),
            system_boot_time=timestamp(uptime_info, "system-booted-time/date-time"),
            protocol_start_time=timestamp(uptime_info, "protocols-started-time/date-time"),
            last_config_time=timestamp(uptime_info, "last-configured-time/date-time"),
            time_source=text(uptime_info, "time-source", strip=True),
        )

    @staticmethod
    async def fetch(device: eznet.device.Device, both_re: bool = False) -> Optional[Dict[str, Uptime]]:
        if not both_re:
            xml = await device.junos.run_xml_cmd("show system uptime")
        else:
            xml = await device.junos.run_xml_cmd("show system uptime invoke-on all-routing-engines")
        if xml is not None:
            system_uptime = xml.find("system-uptime-information")
            if system_uptime is not None:
                return {
                    "localre": Uptime.from_xml(system_uptime)
                }
            mre_results = xml.find("multi-routing-engine-results")
            if mre_results is not None:
                return {
                    text(mre_item, "re-name"):
                        Uptime.from_xml(mre_item.find("system-uptime-information"))
                    for mre_item in mre_results.findall("multi-routing-engine-item")
                    if text(mre_item, "re-name") is not None
                    and mre_item.find("system-uptime-information") is not None
                }
        return None


@define
class CoreDump:
    name: Optional[str]
    size: Optional[int]
    ts: Optional[datetime]
    re: Optional[str] = None
    host: Optional[bool] = None

    @staticmethod
    def from_xml(file_info: _Element) -> CoreDump:
        return CoreDump(
            name=text(file_info, "file-name"),
            size=number(file_info, "file-size"),
            ts=timestamp(file_info, "file-date"),
            re=text(file_info, "../../../re-name"),
            host=file_info.find("../..").attrib.get("style") == "host" or None,
        )

    @staticmethod
    async def fetch(device: eznet.device.Device, both_re: bool = False) -> Optional[List[CoreDump]]:
        if not both_re:
            xml = await device.junos.run_xml_cmd(
                "show system core-dumps",
            )
        else:
            xml = await device.junos.run_xml_cmd(
                "show system core-dumps routing-engine both",
            )
        if xml is not None:
            return [
                CoreDump.from_xml(file_info)
                for file_info in xml.findall(".//directory-list/directory/file-information")
            ]
        return None


@define
class Alarm:
    ts: Optional[datetime]
    cls: Optional[str]
    description: Optional[str]
    type: Optional[str]

    @staticmethod
    def from_xml(alarm: _Element) -> Alarm:
        return Alarm(
            ts=timestamp(alarm, "alarm-time"),
            cls=text(alarm, "alarm-class"),
            description=text(alarm, "alarm-description"),
            type=text(alarm, "alarm-type"),
        )

    @staticmethod
    async def fetch(device: eznet.device.Device) -> Optional[List[Alarm]]:
        xml = await device.junos.run_xml_cmd(
            "show system alarms",
        )
        if xml is not None:
            alarm_info = xml.find("alarm-information")
            if alarm_info is not None:
                return [
                    Alarm.from_xml(alarm)
                    for alarm in alarm_info.findall("alarm-detail")
                ]


class System:
    def __init__(self, device: eznet.device.Device):
        self.info = Data(Info.fetch, device)
        self.sw = Data(SW.fetch, device)
        self.uptime = Data(Uptime.fetch, device)
        self.coredumps = Data(CoreDump.fetch, device)
        self.alarms = Data(Alarm.fetch, device)
        


from typing import ClassVar

from datetime import datetime, timedelta, timezone
import uuid

from attrs import frozen, field
from eznet.db.influxdb import init, Metric, Event

init(
    url="http://172.31.0.33:8086",
    org="lab",
    token="nx0_oCl5srcJNyS7lQiea9FJYhfDrpTniiZpY1AsDeCh1iQGCQqxbA9Ox-_pZYfG-CMFA0xrY3ed1Xm3xAB-OA==",
)


@frozen
class CustomMetric(Metric):
    MEASUREMENT: ClassVar[str] = "custom_metric"

    class Value(Metric.Value):
        value: str
    
    tag: str


@frozen
class CustomEvent(Event):
    MEASUREMENT: ClassVar[str] = "custom_event"

    class Value(Event.Value):
        value: str

    tag: str
    value: Value


def test_metric_write():
    now = datetime.now(timezone.utc)
    CustomMetric(tag="one").write(now, value="one")


def test_metrics_get():
    now = datetime.now(timezone.utc)
    value1 = uuid.uuid4().__str__()
    value2 = uuid.uuid4().__str__()
    CustomMetric(tag="one").write(now, value=value1)
    CustomMetric(tag="two").write(now, value=value2)
    metrics = CustomMetric.get(now-timedelta(microseconds=1), now+timedelta(microseconds=1))
    assert len(metrics) == 2
    assert metrics[CustomMetric(tag="one")][0][0] == now
    assert metrics[CustomMetric(tag="one")][0][1]["value"] == value1
    assert metrics[CustomMetric(tag="two")][0][0] == now
    assert metrics[CustomMetric(tag="two")][0][1]["value"] == value2


def test_metric_read():
    now = datetime.now(timezone.utc)
    value = uuid.uuid4().__str__()
    CustomMetric(tag="one").write(now, value=value)
    CustomMetric(tag="fake").write(now, value="fake")
    metrics = CustomMetric(tag="one").read(now-timedelta(microseconds=1), now+timedelta(microseconds=1))
    assert len(metrics) == 1
    metric = metrics[0]
    assert metric[0] == now
    assert metric[1]["value"] == value


def test_metric_read_first():
    now = datetime.now(timezone.utc)
    value = uuid.uuid4().__str__()
    CustomMetric(tag="one").write(now, value=value)
    CustomMetric(tag="one").write(now+timedelta(microseconds=1), value="fake")
    metrics = CustomMetric(tag="one").read(now-timedelta(microseconds=2), now+timedelta(microseconds=2), first=True)
    assert len(metrics) == 1
    metric = metrics[0]
    assert metric[0] == now
    assert metric[1]["value"] == value


def test_metric_read_last():
    now = datetime.now(timezone.utc)
    value = uuid.uuid4().__str__()
    CustomMetric(tag="one").write(now, value=value)
    CustomMetric(tag="one").write(now+timedelta(microseconds=-1), value="fake")
    metrics = CustomMetric(tag="one").read(now-timedelta(microseconds=2), now+timedelta(microseconds=2), last=True)
    assert len(metrics) == 1
    metric = metrics[0]
    assert metric[0] == now
    assert metric[1]["value"] == value


def test_event_write():
    now = datetime.utcnow()
    CustomEvent(tag="one", timestamp=now, value={"value": "one"}).update(value="one")


def test_events_get():
    now = datetime.utcnow().replace(tzinfo=timezone.utc)
    value1 = uuid.uuid4().__str__()
    value2 = uuid.uuid4().__str__()
    CustomEvent(tag="one", timestamp=now, value={"value": value1}).update(value=value1)
    CustomEvent(tag="two", timestamp=now, value={"value": value2}).update(value=value2)
    metrics = CustomEvent.get(now-timedelta(microseconds=1), now+timedelta(microseconds=1))
    assert len(metrics) == 2
    assert metrics[0].timestamp == now
    assert metrics[1].timestamp == now
#
#
# def test_event_read():
#     now = datetime.utcnow().replace(tzinfo=timezone.utc)
#     value = uuid.uuid4().__str__()
#     CustomEvent(tag="one", timestamp=now).write(value=value)
#     CustomEvent(tag="fake", timestamp=now).write(value="fake")
#     metrics = CustomEvent(tag="one", timestamp=now).read(now-timedelta(microseconds=1), now+timedelta(microseconds=1))
#     assert len(metrics) == 1
#     metric = metrics[0]
#     assert metric[0] == now.replace(tzinfo=timezone.utc)
#     assert metric[1]["value"] == value



import logging
from pathlib import Path
from typing import Union

from rich.console import Console
from rich.highlighter import RegexHighlighter
from rich.logging import RichHandler
from rich.theme import Theme

from eznet.device import Device

MODULE = __name__.split(".")[0]


class LogHighlighter(RegexHighlighter):
    base_style = "eznet."
    highlights = [
        r"(?P<cmd>`.*?`)",
        r"(?P<source>^.*?:)",
        r"(?P<error>ERROR.*)",
    ]


theme = Theme(
    {
        "eznet.cmd": "green",
        "eznet.source": "magenta",
        "eznet.error": "red",
    }
)

console = Console(
    emoji=False,
    markup=False,
    theme=theme,
)


def config_logger(
    level: int,
    file: Union[None, str, Path] = None,
) -> None:
    logger = logging.getLogger(MODULE)
    logger.setLevel(level)

    logger.addHandler(
        RichHandler(
            show_path=False,
            omit_repeated_times=False,
            highlighter=LogHighlighter(),
            console=console,
            level=level,
        )
    )

    if file is not None:
        if isinstance(file, str):
            file = Path(file)
        if not file.parent.exists():
            file.parent.mkdir(parents=True)

        handler = logging.FileHandler(file, mode="w")
        formatter = logging.Formatter(
            "{asctime} {levelname:8s} {message}",
            datefmt="[%x %X]",
            style="{",
        )
        handler.setFormatter(formatter)
        handler.setLevel(level)
        logger.addHandler(handler)


def config_device_logger(
    device: Device,
    level: int,
    file: Union[str, Path],
) -> None:
    if isinstance(file, str):
        file = Path(file)
    if not file.parent.exists():
        file.parent.mkdir(parents=True)

    logger = logging.getLogger(f"{MODULE}.device.{device.id}")
    logger.setLevel(level)

    handler = logging.FileHandler(file, mode="w")
    formatter = logging.Formatter(
        "{asctime} {levelname:8s} {message}",
        datefmt="[%x %X]",
        style="{",
    )
    handler.setFormatter(formatter)
    handler.setLevel(level)
    logger.addHandler(handler)



from __future__ import annotations

import json
import re
from typing import Any, Dict, Optional, Tuple

from lxml import etree
from lxml.etree import _Element  # noqa

from .ssh import SSH

# from .common import CommandError

DEFAULT_CMD_TIMEOUT = 90


class JUNOS:
    def __init__(
        self,
        ssh: SSH,
    ):
        self.ssh = ssh
        self.logger = ssh.logger

    def __str__(self) -> str:
        return f"{self.ssh}: junos"

    def output_has_errors(
        self,
        cmd: str,
        output: str,
    ) -> bool:
        # Junos cli return error in 2nd line of stdout as `error: syntax error, expecting <command>: ...`
        try:
            output_error = output.split("\n")[1].split(": ", 1)
            if output_error[0] == "error":
                self.logger.error(f"{self}: run_cmd `{cmd}`: ERROR: {output_error[1]}")
                # raise CommandError()
                return True
        except IndexError:
            pass

        return False

    async def run_cmd(
        self,
        cmd: str,
        timeout: int = DEFAULT_CMD_TIMEOUT,
    ) -> Optional[str]:
        output, _ = await self.ssh.execute(cmd, timeout=timeout)
        if output is None or self.output_has_errors(cmd, output):
            return None

        return output

    async def run_shell_cmd(
        self,
        cmd: str,
        timeout: int = DEFAULT_CMD_TIMEOUT,
    ) -> Optional[str]:
        output, error = await self.ssh.execute(
            f'start shell command "{cmd}"', timeout=timeout
        )
        # First check for junos error in stdout
        if output is None or self.output_has_errors(cmd, output):
            return None

        # Second check for error in stderr
        if error is not None and error != "":
            self.logger.error(f"{self}: run_shell_cmd: ERROR: {error.strip()}")
            return output

        return output

    async def run_pfe_cmd(
        self,
        cmd: str,
        fpc: int = 0,
        timeout: int = DEFAULT_CMD_TIMEOUT,
    ) -> Optional[str]:
        output, _ = await self.ssh.execute(
            f'request pfe execute target fpc{fpc} command "{cmd}"', timeout=timeout
        )

        # First check for junos error in stdout
        if output is None or self.output_has_errors(cmd, output):
            return None

        # Second check if pfe return error in stdout 2nd line as `Syntax error at 'wrong'`
        try:
            command, error, output = output.split("\n", 2)
            if "error" in error:
                self.logger.error(f"{self}: run_pfe_cmd: ERROR: {error}")
                return None
        except ValueError:
            pass

        return output

    async def run_host_cmd(
        self,
        cmd: str,
        timeout: int = DEFAULT_CMD_TIMEOUT,
    ) -> Optional[str]:
        output, _ = await self.ssh.execute(
            f'request app-engine host-cmd "{cmd}"', timeout=timeout
        )
        # First check for junos error in stdout
        if output is None or self.output_has_errors(cmd, output):
            return None

        # TODO:
        # There is no option to detect error in host shell output :(

        return output

    async def run_xml_cmd(
        self,
        cmd: str,
        timeout: int = DEFAULT_CMD_TIMEOUT,
    ) -> Optional[_Element]:
        output, _ = await self.ssh.execute(f"{cmd} | display xml", timeout=timeout)

        # First check for junos error in stdout
        if output is None or self.output_has_errors(cmd, output):
            return None

        output = output.replace(" xmlns=", " xmlnamespace=").replace("junos:", "")
        try:
            xml = etree.fromstring(output)
        except etree.XMLSyntaxError:
            self.logger.error(f"{self}: run_xml_cmd: xml parse error")
            return None

        # TODO:
        # Verify xml for errors
        # if :
        #     self.logger.error(f"{self}: junos: run_xml_cmd: xml parse error")
        #     return None

        return xml

    async def run_json_cmd(
        self,
        cmd: str,
        timeout: int = DEFAULT_CMD_TIMEOUT,
    ) -> Optional[Dict[Any, Any]]:
        output, _ = await self.ssh.execute(f"{cmd} | display json", timeout=timeout)

        # First check for junos error in stdout
        if output is None or self.output_has_errors(cmd, output):
            return None

        json_output = json.loads(output)
        if not isinstance(json_output, dict):
            self.logger.error(f"{self}: run_json_cmd: json parse error")
            return None

        return json_output

    async def config(
        self,
        config: str,
    ) -> bool:
        if self.ssh.connection is None:
            return False
        self.logger.debug(f"{self}: starting shell")
        stdin, stdout, stderr = await self.ssh.connection.open_session(
            # request_pty=True,
            # term_type='xterm-color',
            # term_size=(80, 24),
        )
        stdin.write("\n")
        while True:
            line = await stdout.readline()
            prompt_match = re.match(r"\w+@[\w.-]+>", line)
            if prompt_match:
                prompt = prompt_match.group(0)[:-1]
                break

        self.logger.debug(f"{self}: ssh shell: got prompt `{prompt}`")

        async def send(cmd: Optional[str] = None) -> Tuple[str, str]:
            if cmd is not None:
                self.logger.debug(f"{self}: ssh shell: sending `{cmd}`")
                stdin.write(cmd + "\n")
            _reply = (await stdout.readuntil(prompt))[0: -len(prompt)]
            self.logger.debug(f"{self}: ssh shell: receive:\n{_reply}")
            _mode = (await stdout.readexactly(2))[0]
            self.logger.debug(f"{self}: ssh shell: mode: `{_mode}`")
            return _reply, _mode

        await send()
        reply, mode = await send("configure private")
        if mode == "#":
            self.logger.info(f"{self}: ssh shell: enter config mode")
            for line in config.split("\n"):
                await send(line)

            reply, mode = await send("commit and-quit")
            if mode == ">":
                self.logger.info(f"{self}: ssh shell: commit successfull")
                return True
            elif mode == "#":
                self.logger.error(
                    f"{self}: ssh shell: commit failed, going to rollback"
                )
                await send("rollback")
                await send("exit")
        else:
            self.logger.error(f"{self}: ssh shell: could not enter to config mode")
        return False



from __future__ import annotations


from typing import Optional, Dict

from attrs import define

from .system import System
from .interfaces import Interface


@define
class Device:
    system: Optional[System] = None
    interfaces: Optional[Dict[str, Interface]] = None



from datetime import datetime
from typing import Optional

from lxml.etree import _Element  # noqa


def text(xml: _Element, xpath: str, strip: bool = False) -> Optional[str]:
    e = xml.find(xpath)
    if e is not None and e.text is not None:
        if not strip:
            return e.text
        else:
            return e.text.strip()
    return None


def number(xml: _Element, xpath: str) -> Optional[int]:
    e = xml.find(xpath)
    if e is not None and e.text is not None and e.text.isdigit():
        return int(e.text)
    return None


def timestamp(xml: _Element, xpath: str) -> Optional[datetime]:
    e = xml.find(xpath)
    if e is not None:
        if e.text is not None and e.text.isdigit():
            return datetime.fromtimestamp(float(e.text))
        e_attrib_seconds = e.attrib.get("seconds")
        if isinstance(e_attrib_seconds, str) and e_attrib_seconds.isdigit():
            try:
                return datetime.fromtimestamp(int(e_attrib_seconds))
            except ValueError:
                pass
        e_attrib_format = e.attrib.get("format")
        if isinstance(e_attrib_format, str):
            try:
                return datetime.strptime(e_attrib_format, "%b %d %Y")
            except ValueError:
                pass
    return None



from __future__ import annotations

import asyncio
import logging
import os
import socket
from enum import Enum, auto
from pathlib import Path
from time import time
from types import TracebackType
from typing import ClassVar, Dict, List, Optional, Tuple, Type, Union

import asyncssh

DEFAULT_CONNECT_TIMEOUT = 30
DEFAULT_CMD_TIMEOUT = 90
DEFAULT_KEEPALIVE = 5
DEFAULT_ENCODING = "latin-1"

MAX_SIMULTANEOUS_CONNECTIONS = 64
MAX_SIMULTANEOUS_EXECUTIONS = 64
MAX_SIMULTANEOUS_DOWNLOADS = 4
MAX_SIMULTANEOUS_UPLOADS = 4


class ConnectError(Exception):
    pass


class AuthenticationError(Exception):
    pass


class State(Enum):
    DISCONNECTED = auto()
    CONNECTING = auto()
    CONNECTED = auto()
    WAITING_CONNECT = auto()
    WAITING_RECONNECT = auto()

    def __repr__(self) -> str:
        return self.name

    def __str__(self) -> str:
        return self.name


class SSH:
    def __init__(
        self,
        ip: Optional[str],
        user_name: Optional[str] = None,
        user_pass: Optional[str] = None,
        device: Optional[str] = None,
    ) -> None:
        self.ip = ip
        self.user_name = user_name or os.environ["USER"]
        self.user_pass = user_pass
        self.device = device

        if device is None:
            self.logger = logging.getLogger("eznet.device")
        else:
            self.logger = logging.getLogger(f"eznet.device.{device}")

        self.connection: Optional[asyncssh.SSHClientConnection] = None
        self.state = State.DISCONNECTED
        self.error: Optional[str] = None

        self.requests: List[Request] = []

    def __str__(self) -> str:
        if self.device is not None:
            return f"{self.device} (ip={self.ip or ''}): ssh"
        else:
            return f"{self.ip or super().__str__()}: ssh"

    async def __aenter__(self) -> None:
        await self.connect()

    async def __aexit__(
        self,
        exc_type: Optional[Type[BaseException]],
        exc: Optional[BaseException],
        traceback: Optional[TracebackType],
    ) -> None:
        self.disconnect()

    async def connect(self, attempts: int = 1, attempt_timeout: int = 15) -> None:
        if self.ip is None:
            self.logger.warning(f"{self}: ip is not set")
            raise ConnectError()

        async with Lock.get(self).connect:
            if self.connection is not None:
                return

            self.state = State.WAITING_CONNECT
            await Semaphore.get().connect.acquire()
            while attempts > 0:
                self.state = State.CONNECTING
                self.logger.info(
                    f"{self}: connecting to {self.ip} as {self.user_name}"
                )
                try:
                    self.connection = await asyncssh.connect(
                        host=self.ip,
                        username=self.user_name,
                        password=self.user_pass,
                        client_factory=create_client_factory(self),
                        connect_timeout=DEFAULT_CONNECT_TIMEOUT,
                        keepalive_interval=DEFAULT_KEEPALIVE,
                        known_hosts=None,
                    )
                except (
                    socket.gaierror,
                    TimeoutError,
                    asyncio.exceptions.TimeoutError,
                    ConnectionRefusedError,
                    ConnectionResetError,
                    OSError,  # Network unreachable
                    asyncssh.Error,
                ) as err:
                    self.state = State.DISCONNECTED
                    self.error = f"{err.__class__.__name__}"
                    self.logger.error(f"{self}: {err.__class__.__name__}: {err}")
                    # Semaphore.get().connect.release()
                    # raise ConnectError() from None
                except Exception as err:
                    self.state = State.DISCONNECTED
                    self.error = f"{err.__class__.__name__}"
                    self.logger.critical(f"{self}: {err.__class__.__name__}: {err}")
                    Semaphore.get().connect.release()
                    raise
                else:
                    self.state = State.CONNECTED
                    self.error = None
                    self.logger.info(f"{self}: CONNECTED")
                    return

                attempts -= 1
                if attempts > 0:
                    self.state = State.WAITING_RECONNECT
                    await asyncio.sleep(attempt_timeout)

            Semaphore.get().connect.release()
            # raise ConnectError() from None

    def disconnect(self) -> None:
        if self.connection is not None:
            self.connection.close()

    async def execute(
        self, cmd: str, timeout: int = DEFAULT_CMD_TIMEOUT
    ) -> Tuple[Optional[str], Optional[str]]:
        if self.connection is not None:
            request = CmdRequest(cmd)
            self.requests.append(request)

            try:
                chan, session = await self.connection.create_session(
                    create_session_factory(request), cmd, encoding=DEFAULT_ENCODING,
                )
                self.logger.info(f"{self}: execute `{cmd}`: waiting for reply")
                # TODO: add watchguard
                # await chan.wait_closed()
                await asyncio.wait_for(chan.wait_closed(), timeout=timeout)
            except (
                asyncio.TimeoutError,
                asyncssh.Error,
            ) as err:
                self.logger.error(
                    f"{self}: execute `{cmd}`: {err.__class__.__name__}: {err}"
                )
                self.error = f"{err.__class__.__name__}"
                # raise CommandError()
                return None, None
            else:
                self.logger.info(
                    f"{self}: execute `{cmd}`: "
                    f"got reply: {len(request.stdout)} bytes / {len(request.stderr)} bytes"
                )
                if request.stdout:
                    self.logger.debug(
                        f"{self}: execute `{cmd}`: stdout:\n{request.stdout}"
                    )
                if request.stderr:
                    self.logger.debug(
                        f"{self}: execute `{cmd}`: stderr:\n{request.stderr}"
                    )

                self.requests.remove(request)
                return request.stdout, request.stderr

        self.logger.warning(f"{self}: execute `{cmd}` no active connection")
        # raise CommandError()
        return None, None

    async def download(self, src: str, dst: Union[str, Path]) -> List[str]:
        download_files: List[str] = []

        async with Semaphore.get().download:
            request = FileRequest(src)
            t0 = t1 = time()
            r1 = 0

            def progress_handler(
                src_file: bytes, dst_file: bytes, received: int, total: int
            ) -> None:
                if dst_file.decode(DEFAULT_ENCODING) not in download_files:
                    download_files.append(dst_file.decode(DEFAULT_ENCODING))
                nonlocal t0, t1, r1, request

                if request.file_name != src_file.decode(DEFAULT_ENCODING):
                    self.requests.remove(request)
                    request = FileRequest(src_file.decode(DEFAULT_ENCODING))
                    self.requests.append(request)

                request.received_bytes = received
                request.total_bytes = total

                t_delta = time() - t1
                if received == total:
                    t_delta = time() - t0
                    received_part = received / total if total > 0 else 1
                    speed = received / t_delta
                    self.logger.info(
                        f"{self}: download `{src_file.decode('ascii')}`: {received:,} of {total:,}:"
                        f" {received_part:.0%} at {speed:,.0f} Bps"
                    )
                    t0 = t1 = time()
                    r1 = 0
                    request.speed = speed
                elif t_delta > 10:
                    received_part = received / total if total > 0 else 1
                    speed = (received - r1) / t_delta
                    self.logger.info(
                        f"{self}: downloading `{src_file.decode('ascii')}`: {received:,} of {total:,}:"
                        f" {received_part:.0%} at {speed:,.0f} Bps"
                    )
                    t1 = time()
                    r1 = received
                    request.speed = speed

            try:
                self.requests.append(request)
                await asyncssh.scp(
                    (self.connection, src),
                    dst,
                    progress_handler=progress_handler,
                    preserve=True,
                    recurse=True,
                )
            except (
                asyncssh.SFTPError,
                asyncssh.SFTPFailure,
            ) as err:
                self.logger.error(
                    f"{self}: download `{src}` --> `{dst}`: {err.__class__.__name__}: {err}"
                )
            else:
                self.logger.info(f"{self}: download `{src}` --> `{dst}`: DONE")
            finally:
                self.requests.remove(request)
                return download_files

    async def upload(self, src: Union[str, Path], dst: str) -> List[str]:
        upload_files: List[str] = []

        async with Semaphore.get().upload:
            request = FileRequest(src)
            t0 = t1 = time()
            r1 = 0

            def progress_handler(
                src_file: bytes, dst_file: bytes, received: int, total: int
            ) -> None:
                if dst_file.decode(DEFAULT_ENCODING) not in upload_files:
                    upload_files.append(dst_file.decode(DEFAULT_ENCODING))
                nonlocal t0, t1, r1, request

                if request.file_name != src_file.decode(DEFAULT_ENCODING):
                    self.requests.remove(request)
                    request = FileRequest(src_file.decode(DEFAULT_ENCODING))
                    self.requests.append(request)

                request.received_bytes = received
                request.total_bytes = total

                t_delta = time() - t1
                if received == total:
                    t_delta = time() - t0
                    received_part = received / total if total > 0 else 1
                    speed = received / t_delta
                    self.logger.info(
                        f"{self}: upload `{src_file.decode('ascii')}`: {received:,} of {total:,}:"
                        f" {received_part:.0%} at {speed:,.0f} Bps"
                    )
                    t0 = t1 = time()
                    r1 = 0
                    request.speed = speed
                elif t_delta > 10:
                    received_part = received / total if total > 0 else 1
                    speed = (received - r1) / t_delta
                    self.logger.info(
                        f"{self}: uploading `{src_file.decode('ascii')}`: {received:,} of {total:,}:"
                        f" {received_part:.0%} at {speed:,.0f} Bps"
                    )
                    t1 = time()
                    r1 = received
                    request.speed = speed

            try:
                self.requests.append(request)
                await asyncssh.scp(
                    src,
                    (self.connection, dst),
                    progress_handler=progress_handler,
                    preserve=True,
                    recurse=True,
                )
            except (
                asyncssh.SFTPError,
                asyncssh.SFTPFailure,
            ) as err:
                self.logger.error(
                    f"{self}: upload `{src}` --> `{dst}`: {err.__class__.__name__}: {err}"
                )
            else:
                self.logger.info(f"{self}: upload `{src}` --> `{dst}`: DONE")
            finally:
                self.requests.remove(request)
                return upload_files


def create_client_factory(ssh: SSH) -> Type[asyncssh.SSHClient]:
    class SSHClient(asyncssh.SSHClient):
        def connection_lost(self, err: Optional[Exception]) -> None:
            if ssh.connection is not None:
                ssh.connection = None
                ssh.state = State.DISCONNECTED
                Semaphore.get().connect.release()
                if err is None:
                    ssh.logger.info(f"{ssh}: DISCONNECTED")
                else:
                    ssh.logger.error(f"{ssh}: DISCONNECTED: {err}")
                    ssh.error = f"{err.__class__.__name__}"
                    # ssh.status = Status.ERROR

                    # try:
                    #     loop = asyncio.get_running_loop()
                    #     loop.create_task(ssh.connect(
                    #         attempts=RECONNECT_ATTEMPTS,
                    #         attempt_timeout=RECONNECT_ATTEMPT_TIMEOUT,
                    #     ))
                    # except RuntimeError as err:
                    #     ssh.logger.critical(f"{ssh}: reconnect error: {err}")

    return SSHClient


def create_session_factory(request: CmdRequest) -> Type[asyncssh.SSHClientSession[str]]:
    class SSHClientSession(asyncssh.SSHClientSession[str]):
        def data_received(self, data: str, datatype: asyncssh.DataType) -> None:
            if datatype == asyncssh.EXTENDED_DATA_STDERR:
                request.stderr += data
            else:
                request.stdout += data

    return SSHClientSession


class Request:
    pass


class CmdRequest(Request):
    def __init__(self, cmd: str):
        self.cmd = cmd
        self.stdout = ""
        self.stderr = ""

    def __repr__(self) -> str:
        return f"{self.cmd}\t{len(self.stdout):,}\t/\t{len(self.stderr):,}"


class FileRequest(Request):
    def __init__(self, file_name: str, received_bytes: int = 0, total_bytes: int = 0):
        self.file_name = file_name
        self.received_bytes = received_bytes
        self.total_bytes = total_bytes
        self.speed: float = 0

    def __repr__(self) -> str:
        received_part = (
            self.received_bytes / self.total_bytes if self.total_bytes > 0 else 1
        )
        return (
            f"{self.file_name}\t"
            f"{self.received_bytes:,}\tof\t{self.total_bytes:,}\t"
            f"[ {received_part:.0%} ]\t"
            f"at {self.speed:,.0f} Bps"
        )


class Lock:
    loop: ClassVar[Optional[asyncio.AbstractEventLoop]] = None
    instances: ClassVar[Dict[int, Lock]] = {}

    def __init__(self) -> None:
        self.connect = asyncio.Lock()

    @classmethod
    def get(cls, ssh: SSH) -> Lock:
        loop = asyncio.get_running_loop()
        if Lock.loop != loop:
            Lock.loop = loop
            Lock.instances = {}
        if id(ssh) not in Lock.instances:
            Lock.instances[id(ssh)] = Lock()
        return Lock.instances[id(ssh)]


class Semaphore:
    loop: ClassVar[Optional[asyncio.AbstractEventLoop]] = None
    instances: ClassVar[Optional[Semaphore]] = None

    def __init__(self) -> None:
        self.connect = asyncio.Semaphore(MAX_SIMULTANEOUS_CONNECTIONS)
        self.execute = asyncio.Semaphore(MAX_SIMULTANEOUS_EXECUTIONS)
        self.download = asyncio.Semaphore(MAX_SIMULTANEOUS_DOWNLOADS)
        self.upload = asyncio.Semaphore(MAX_SIMULTANEOUS_UPLOADS)

    @classmethod
    def get(cls) -> Semaphore:
        loop = asyncio.get_running_loop()
        if Semaphore.loop != loop or Semaphore.instances is None:
            Semaphore.loop = loop
            Semaphore.instances = Semaphore()
        return Semaphore.instances



from __future__ import annotations

from typing import Optional, Dict

from attrs import define


@define
class Peer:
    device: Optional[str] = None
    interface: Optional[str] = None


@define
class Member:
    peer: Optional[Peer] = None


@define
class Interface:
    members: Optional[Dict[str, Member]] = None
    peer: Optional[Peer] = None



import inspect
from types import FrameType
from typing import Optional, Type, TypeVar

C = TypeVar("C")


def get_caller_object(cls: Type[C], frame: Optional[FrameType] = None) -> Optional[C]:
    frame = frame or inspect.currentframe()
    if frame is None:
        return None

    if "self" in frame.f_locals and isinstance(frame.f_locals["self"], cls):
        return frame.f_locals["self"]
    else:
        frame = frame.f_back
    if frame is None:
        return None
    else:
        return get_caller_object(cls, frame)



from __future__ import annotations

from pathlib import Path
from typing import Union, Iterable, List

from eznet.device import Device


def cli_commands(device: Device) -> Iterable[str]:
    commands: List[str] = [
        "show system information",
        "show version",
        "show version invoke-on all-routing-engines",
        "show system uptime",
        "show system uptime invoke-on all-routing-engines",
        "show system alarms",
        "show system process",
        "show system memory",
        "show system core-dumps",
        "show system core-dumps routing-engine both",
        "show system core-dumps all-members",
        "show system core-dumps satellite",
        "show krt queue",
        "show chassis alarms",
        "show chassis alarms satellite",
        "show chassis hardware",
        "show chassis hardware models",
        "show chassis firmware",
        "show chassis routing-engine",
        "show chassis fpc",
        "show chassis fpc pic-status",
        *[
            f"show chassis pic fpc-slot {fpc_number} pic-slot {pic_number}"
            for fpc_number, fpc in device.info.chassis.fpc["default"].items()
            for pic_number in fpc.pics.keys()
        ],
        "show chassis satellite",
        "show chassis satellite detail",
        "show chassis satellite software",
        "show chassis satellite extended-port",
        "show interfaces summary",
        "show interfaces",
        "show lacp interfaces",
        "show lldp detail",
        "show lldp neighbors",
        "show route summary",
        "show route forwarding-table summary",
        "show route cumulative vpn-family inet.0",
        "show route cumulative vpn-family inet6.0",
        "show bgp summary",
        "show bfd session",
    ]
    yield from commands


def pfe_commands(device: Device) -> Iterable[str]:
    commands: List[str] = [
        "show heap",
        "show syslog messages",
    ]
    yield from commands


def shell_commands(device: Device) -> Iterable[str]:
    commands: List[str] = [
        "ls -l",
        "ls -l /var/db/scripts/op",
        "ls -l /var/db/scripts/event",
        "ls -l /var/db/scripts/commit",
    ]
    yield from commands


def host_commands(device: Device) -> Iterable[str]:
    # commands: List[str] = [
    #     *([
    #         "date +'%Y-%m-%d %H:%M:%S'",
    #         "ps -elf",
    #         "free -m",
    #       ] if device.info.system.info().sw_family in ["junos-qfx"] else []),
    # ]
    commands: List[str] = []
    yield from commands


def log_files(device: Device) -> Iterable[str]:
    files: List[str] = [
        # "/var/log/*",
        "/var/log/messages*",
        "/var/log/chassis*",
        "/var/log/authorization-commands*",
    ]
    yield from files


async def rsi(
    device: Device, job_path: Union[Path, str, None],
):
    if isinstance(job_path, str):
        job_path = Path(job_path)

    with open(job_path / f"{device.id}.info", "w") as io:
        print(await device.info.system.info(), file=io)
        print(await device.info.chassis.re(), file=io)
        print(await device.info.chassis.fpc(), file=io)
        print(await device.info.system.uptime(), file=io)

    if not job_path.exists():
        job_path.mkdir(parents=True)

    with open(job_path / f"{device.id}.cmd", "w") as io:
        for cmd in cli_commands(device):
            print(f"{' ' + cmd + ' ':=^120}", file=io)
            output = await device.junos.run_cmd(cmd)
            if output is not None:
                print(output, file=io)
            print(f"{' ' + cmd + ' ':^^120}", file=io)
            print(file=io)

        for cmd in shell_commands(device):
            print(f"{' ' + cmd + ' ':=^120}", file=io)
            output = await device.junos.run_shell_cmd(cmd)
            if output is not None:
                print(output, file=io)
            print(f"{' ' + cmd + ' ':^^120}", file=io)
            print(file=io)

    for fpc_number in device.info.chassis.fpc["default"].keys():
        with open(job_path / f"{device.id}.fpc{fpc_number}", "w") as io:
            for cmd in pfe_commands(device):
                print(f"{' ' + cmd + ' ':=^120}", file=io)
                output = await device.junos.run_pfe_cmd(cmd, fpc=fpc_number)
                if output is not None:
                    print(output, file=io)
                print(f"{' ' + cmd + ' ':^^120}", file=io)
                print(file=io)

    device_job_path = job_path / f"{device.id}"
    if not device_job_path.exists():
        device_job_path.mkdir(parents=True)
    await download(device, "/var/log", device_job_path)

    # for file in log_files(device):
    #     remote_path = Path(file)
    #     if remote_path.is_absolute():
    #         local_path = job_path / f"{device.id}" / remote_path.parent.relative_to("/")
    #     else:
    #         local_path = job_path / f"{device.id}" / remote_path.parent
    #     if not local_path.exists():
    #         local_path.mkdir(parents=True)
    #     await device.ssh.download(file, local_path)
    #
    # with open(job_path / f"{device.id}.rsi", "w") as io:
    #     output = await device.junos.run_cmd("request support information", timeout=600)
    #     if output is not None:
    #         print(output, file=io)


async def download(device: Device, remote_path: Union[Path, str], local_path: Union[Path, str]):
    if isinstance(remote_path, str):
        remote_path = Path(remote_path)
    if remote_path.is_absolute():
        tmp_file_name = (
            f"{Path(remote_path).relative_to('/')}"
            .replace("/", ".")
            .replace("*", "")
            + ".tgz"
        )
    else:
        tmp_file_name = (
            f"{Path(remote_path)}"
            .replace("/", ".")
            .replace("*", "")
            + ".tgz"
        )
    await device.junos.run_cmd(
        f'request routing-engine execute command "tar -czf ./{tmp_file_name} {remote_path}" routing-engine both'
    )
    await device.junos.run_cmd(f"file copy re0:./{tmp_file_name} ./re0.{tmp_file_name}")
    await device.junos.run_cmd(f"file copy re1:./{tmp_file_name} ./re1.{tmp_file_name}")
    await device.ssh.download(f"re0.{tmp_file_name}", local_path)
    await device.ssh.download(f"re1.{tmp_file_name}", local_path)
    await device.junos.run_cmd(f"file delete ./re0.{tmp_file_name}")
    await device.junos.run_cmd(f"file delete ./re1.{tmp_file_name}")


def arch_cli_commands(device: Device) -> Iterable[str]:
    commands: List[str] = [
        "file archive compress source /var/log destination /var/tmp/var.log.tgz",
        "file archive source /config/juniper.conf.*.gz destination /var/tmp/config.tar",
        "file archive source /var/db/config/juniper.conf.*.gz destination /var/tmp/var.db.config.tar",
    ]
    yield from commands


def arch_host_commands(device: Device) -> Iterable[str]:
    commands: List[str] = [
        "tar cvfz /var/tmp/hostvar.log.tgz /var/log",
        "mkdir -p /var/tmp/dcpfe",
        "scp -o StrictHostKeyChecking=no 192.168.1.16:/var/log/*.log /var/tmp/dcpfe",
        "tar cvfz /var/tmp/dcpfe.log.tgz /var/tmp/dcpfe",
        "rm -rf /var/tmp/dcpfe",
    ]
    yield from commands


def arch_files(device: Device) -> Iterable[str]:
    files: List[str] = [
        # "~/rsi.txt",
        "/var/tmp/var.log.tgz",
        "/var/tmp/config.tar",
        "/var/tmp/var.db.config.tar",
        "/hostvar/tmp/hostvar.log.tgz",
        "/hostvar/tmp/dcpfe.log.tgz",
    ]
    yield from files



from __future__ import annotations

from datetime import datetime
from typing import Dict, Any, Optional, List

from attrs import define
from lxml.etree import _Element

from eznet.parser.xml import text, number, timestamp
import eznet


@define
class AF:
    address: Optional[str]
    network: Optional[str]

    @staticmethod
    def from_xml(xml: _Element) -> AF:
        return AF(
            address=text(xml, "interface-address/ifa-local"),
            network=text(xml, "interface-address/ifa-destination"),
        )


@define
class UnitTraffic:
    input_packets: Optional[int]
    output_packets: Optional[int]

    @staticmethod
    def from_xml(xml: _Element) -> UnitTraffic:
        return UnitTraffic(
            input_packets=number(xml, "traffic-statistics/input-packets"),
            output_packets=number(xml, "traffic-statistics/output-packets"),
        )


@define
class Unit:
    description: str
    family: Dict[str, AF]
    traffic: UnitTraffic

    @staticmethod
    def from_xml(xml: _Element) -> Unit:
        return Unit(
            description=text(xml, "description"),
            family={
                text(e, "address-family-name"): AF.from_xml(e)
                for e in xml.findall("address-family")
            },
            traffic=UnitTraffic.from_xml(xml),
        )


@define
class Traffic:
    input_bps: int
    input_pps: int
    output_bps: int
    output_pps: int

    @staticmethod
    def from_xml(xml: _Element) -> Traffic:
        return Traffic(
            input_bps=number(xml, "traffic-statistics/input-bps"),
            input_pps=number(xml, "traffic-statistics/input-pps"),
            output_bps=number(xml, "traffic-statistics/output-bps"),
            output_pps=number(xml, "traffic-statistics/output-pps"),
        )


@define
class Interface:
    admin: Optional[str]
    oper: Optional[str]
    description: Optional[str]
    speed: Optional[str]
    units: Dict[int, Unit]
    traffic: Traffic
    ae: Optional[str]

    @staticmethod
    def from_xml(xml: _Element) -> Interface:
        ae = text(xml, "logical-interface/address-family[address-family-name=\"aenet\"]ae-bundle-name")
        if ae is not None:
            ae = ae.split(".")[0]
        return Interface(
            description=text(xml, "description"),
            admin=text(xml, "admin-status"),
            oper=text(xml, "oper-status"),
            speed=text(xml, "speed"),
            units={
                int(text(e, "name").split(".")[1]): Unit.from_xml(e)
                for e in xml.findall("logical-interface")
            },
            traffic=Traffic.from_xml(xml),
            ae=ae,
        )

    @classmethod
    async def fetch(cls, device: eznet.device.Device) -> Optional[Dict[str, Interface]]:
        show_interfaces = await device.junos.run_xml_cmd("show interfaces")
        if show_interfaces is not None:
            interface_information = show_interfaces.find("interface-information")
            return {
                text(e, "name"): Interface.from_xml(e)
                for e in interface_information.findall("physical-interface")
                if text(e, "name")[:2] in ["ae", "ge", "xe", "et"]
            }



import asyncio
from contextlib import suppress


async def process(i):
    await asyncio.sleep(i)
    print(i)
    with suppress(Exception):
        if i == 1:
            raise Exception(i)
    return i


async def run():
    ret = await asyncio.gather(*(process(i) for i in range(5)), return_exceptions=True)
    print(ret)
    print(any(map(lambda x: isinstance(x, Exception), ret)))
    print(all(map(lambda x: isinstance(x, Exception), ret)))

asyncio.run(run())



import pathlib

from setuptools import find_packages, setup

pwd = pathlib.Path(__file__).parent
requirements_file = pathlib.Path(pwd, "requirements.txt")
install_requires = requirements_file.read_text().splitlines()


setup(
    name="eznet",
    version="0.0.1",
    packages=find_packages(),
    url="https://github.com/omaxx/eznet",
    license="",
    author="maxx orlov",
    author_email="",
    description="",
    entry_points={
        "console_scripts": "eznet=eznet.__main__:cli",
    },
    install_requires=install_requires,
)



from __future__ import annotations

import logging
from typing import Dict, List, Optional, Union, Any

import cattrs

from . import info
from . import vars
from .drivers import JUNOS, SSH

converter = cattrs.Converter()


class Device:
    def __init__(
        self,
        name: str,
        site: Optional[str] = None,
        ip: Union[str, List[str], Dict[str, str], None] = None,
        user_name: Optional[str] = None,
        user_pass: Optional[str] = None,
        root_pass: Optional[str] = None,
        **kwargs,
    ):
        self.name = name
        self.site = site
        self.id = f"{self.site}.{self.name}" if self.site is not None else self.name

        self.ip = ip or name
        self.user_name = user_name
        self.user_pass = user_pass
        self.root_pass = root_pass
        self.kwargs = kwargs

        self.logger = logging.getLogger(f"eznet.device.{self.id}")

        self.vars: Optional[vars.Device] = None

        if isinstance(ip, str):
            ssh_ip = ip
        elif isinstance(ip, list) and len(ip) > 0:
            ssh_ip = ip[0]
        elif isinstance(ip, dict) and len(ip) > 0:
            ssh_ip = list(ip.values())[0]
        else:
            ssh_ip = None

        self.ssh = SSH(
            ip=ssh_ip,
            user_name=user_name,
            user_pass=user_pass,
            device=self.id,
        )
        self.junos = JUNOS(
            ssh=self.ssh,
        )

        self.info = info.Device(self)

    def __str__(self) -> str:
        return self.id

    def __repr__(self) -> str:
        return f"Device(id={self.id})"

    def __eq__(self, other: Any) -> bool:
        return isinstance(other, Device) and self.id == other.id

    def import_vars(self, **data):
        self.vars = converter.structure(data, vars.Device)



from __future__ import annotations

from typing import Generic, TypeVar, Dict, Callable, Awaitable, Any, get_type_hints, Optional
from typing_extensions import ParamSpec, Concatenate

import cattrs

import eznet


DEFAULT_TAG = "default"

converter = cattrs.Converter()

P = ParamSpec('P')
V = TypeVar("V")


class Data(Generic[V]):
    def __init__(
        self,
        func: Callable[Concatenate[eznet.device.Device, P], Awaitable[Optional[V]]],
        device: eznet.device.Device,
    ) -> None:
        self.cls = get_type_hints(func)['return']
        self.data: Dict[str, V] = {}
        self.func = func
        self.device = device

    def __getitem__(self, tag: str) -> V:
        return self.data[tag]

    def imp0rt(self, data: Any, tag: str = DEFAULT_TAG) -> None:
        self.data[tag] = converter.structure(data, self.cls)

    @property
    def v(self) -> Optional[V]:
        return self.data.get(DEFAULT_TAG)

    async def fetch(self, tag: str = DEFAULT_TAG, *args: P.args, **kwargs: P.kwargs) -> Optional[V]:
        data = await self.func(self.device, *args, **kwargs)
        if data is not None:
            self.data[tag] = data
            return self.data[tag]
        return None

    def exp0rt(self, tag: str = DEFAULT_TAG) -> Any:
        data = converter.unstructure(self.data[tag])
        return data



#!/usr/bin/env python3

from __future__ import annotations

import asyncio
import logging
from datetime import datetime
import sys
from typing import Optional, Union, Tuple
from pathlib import Path

import click
from rich import print

from eznet.logger import config_logger, config_device_logger
from eznet.inventory import Inventory
from eznet.device import Device
from eznet.rsi import rsi

JOB_TS_FORMAT = "%Y%m%d-%H%M%S"


def set_path(
    work_path: Union[Path, str],
    jobs_path: Union[Path, str],
    job_name: str,
) -> Tuple[Path, Path, Path, Path]:
    if isinstance(work_path, str):
        work_path = Path(work_path)
    work_path.expanduser()

    if isinstance(jobs_path, str):
        jobs_path = Path(jobs_path)
    jobs_path.expanduser()
    jobs_path = work_path / jobs_path

    job_path = jobs_path / job_name
    log_file = jobs_path / f"{job_name}.log"

    return work_path, jobs_path, job_path, log_file


def main(
    inventory_path: Union[Path, str],
    work_path: Union[Path, str] = "working",
    jobs_path: Union[Path, str] = "jobs",
    job_name: Optional[str] = None,
    device_id: Optional[str] = None,
) -> None:
    time_start = datetime.now()
    job_name = job_name or time_start.strftime(JOB_TS_FORMAT)
    print(f"{job_name}: [black on white]job started at {time_start}")

    work_path, jobs_path, job_path, log_file = set_path(work_path, jobs_path, job_name)
    config_logger(logging.INFO, log_file)
    print(f"{job_name}: [black on white]log file: {log_file.absolute()}, job folder: {job_path.absolute()}")

    try:
        inventory = Inventory()
        inventory.load(inventory_path)

        def device_filter(device):
            return device_id is None or device.id == device_id

        async def process(
            device: Device,
        ) -> None:
            config_device_logger(device, logging.DEBUG, file=job_path / f"{device.id}.log")

            async with device.ssh:
                # TODO: print status
                await device.info.system.info.fetch()
                await device.info.chassis.re.fetch()
                await device.info.chassis.fpc.fetch()
                await device.info.system.uptime.fetch()

                await rsi(device, job_path=job_path)

            if device.ssh.error is not None:
                # TODO: print error status
                pass

        async def gather() -> None:
            await asyncio.gather(*(
                process(device) for device in inventory.devices if device_filter(device)
            ), return_exceptions=False)

        asyncio.run(gather())
    except KeyboardInterrupt:
        print(f"{job_name}: [white on red]keyboard interrupted")
        sys.exit(130)
    finally:
        time_stop = datetime.now()
        print(f"{job_name}: [black on white]job finished at {time_stop}")
        print(f"{job_name}: [black on white]log file: {log_file.absolute()}, job folder: {job_path.absolute()}")


@click.command
@click.option(
    "--inventory", "-i", "inventory_path",
    help="Inventory path", type=click.Path(exists=True),
    default="inventory/devices", show_default=True,
)
@click.option(
    "--work", "-w", "work_path",
    help="Working path", type=click.Path(),
    default="working", show_default=True,
)
@click.option(
    "--jobs", "-j", "jobs_path",
    help="Jobs path (absolute or related to working path)", type=click.Path(),
    default="jobs", show_default=True,
)
@click.option(
    "--device-id", "-d", "device_id",
    help="Device filter: id",
)
def cli(**kwargs) -> None:
    main(**kwargs)


if __name__ == "__main__":
    cli()
